const char error_lua[] =
"-- error.lua (internal file)\n"
"\n"
"local ffi = require('ffi')\n"
"ffi.cdef[[\n"
"struct type_info;\n"
"\n"
"enum {\n"
"    DIAG_ERRMSG_MAX = 512,\n"
"    DIAG_FILENAME_MAX = 256\n"
"};\n"
"\n"
"typedef void (*error_f)(struct error *e);\n"
"\n"
"struct error {\n"
"    error_f _destroy;\n"
"    error_f _raise;\n"
"    error_f _log;\n"
"    const struct type_info *_type;\n"
"    int _refs;\n"
"    /** Line number. */\n"
"    unsigned _line;\n"
"    /* Source file name. */\n"
"    char _file[DIAG_FILENAME_MAX];\n"
"    /* Error description. */\n"
"    char _errmsg[DIAG_ERRMSG_MAX];\n"
"};\n"
"\n"
"char *\n"
"exception_get_string(struct error *e, const struct method_info *method);\n"
"int\n"
"exception_get_int(struct error *e, const struct method_info *method);\n"
"]]\n"
"\n"
"local REFLECTION_CACHE = {}\n"
"\n"
"local function reflection_enumerate(err)\n"
"    local key = tostring(err._type)\n"
"    local result = REFLECTION_CACHE[key]\n"
"    if result ~= nil then\n"
"        return result\n"
"    end\n"
"    result = {}\n"
"    -- See type_foreach_method() in reflection.h\n"
"    local t = err._type\n"
"    while t ~= nil do\n"
"        local m = t.methods\n"
"        while m.name ~= nil do\n"
"            result[ffi.string(m.name)] = m\n"
"            m = m + 1\n"
"        end\n"
"        t = t.parent\n"
"    end\n"
"    REFLECTION_CACHE[key] = result\n"
"    return result\n"
"end\n"
"\n"
"local function reflection_get(err, method)\n"
"    if method.nargs ~= 0 then\n"
"        return nil -- NYI\n"
"    end\n"
"    if method.rtype == ffi.C.CTYPE_INT then\n"
"        return tonumber(ffi.C.exception_get_int(err, method))\n"
"    elseif method.rtype == ffi.C.CTYPE_CONST_CHAR_PTR then\n"
"        local str = ffi.C.exception_get_string(err, method)\n"
"        if str == nil then\n"
"            return nil\n"
"        end\n"
"        return ffi.string(str)\n"
"    end\n"
"end\n"
"\n"
"local function error_type(err)\n"
"    return ffi.string(err._type.name)\n"
"end\n"
"\n"
"local function error_message(err)\n"
"    return ffi.string(err._errmsg)\n"
"end\n"
"\n"
"local function error_trace(err)\n"
"    if err._file[0] == 0 then\n"
"        return {}\n"
"    end\n"
"    return {\n"
"        { file = ffi.string(err._file), line = tonumber(err._line) };\n"
"    }\n"
"end\n"
"\n"
"local error_fields = {\n"
"    [\"type\"]        = error_type;\n"
"    [\"message\"]     = error_message;\n"
"    [\"trace\"]       = error_trace;\n"
"}\n"
"\n"
"local function error_unpack(err)\n"
"    if not ffi.istype('struct error', err) then\n"
"        error(\"Usage: error:unpack()\")\n"
"    end\n"
"    local result = {}\n"
"    for key, getter in pairs(error_fields)  do\n"
"        result[key] = getter(err)\n"
"    end\n"
"    for key, getter in pairs(reflection_enumerate(err)) do\n"
"        local value = reflection_get(err, getter)\n"
"        if value ~= nil then\n"
"            result[key] = value\n"
"        end\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"local function error_raise(err)\n"
"    if not ffi.istype('struct error', err) then\n"
"        error(\"Usage: error:raise()\")\n"
"    end\n"
"    error(err)\n"
"end\n"
"\n"
"local function error_match(err, ...)\n"
"    if not ffi.istype('struct error', err) then\n"
"        error(\"Usage: error:match()\")\n"
"    end\n"
"    return string.match(error_message(err), ...)\n"
"end\n"
"\n"
"local function error_serialize(err)\n"
"    -- Return an error message only in admin console to keep compatibility\n"
"    return error_message(err)\n"
"end\n"
"\n"
"local error_methods = {\n"
"    [\"unpack\"] = error_unpack;\n"
"    [\"raise\"] = error_raise;\n"
"    [\"match\"] = error_match; -- Tarantool 1.6 backward compatibility\n"
"    [\"__serialize\"] = error_serialize;\n"
"}\n"
"\n"
"local function error_index(err, key)\n"
"    local getter = error_fields[key]\n"
"    if getter ~= nil then\n"
"        return getter(err)\n"
"    end\n"
"    getter = reflection_enumerate(err)[key]\n"
"    if getter ~= nil and getter.nargs == 0 then\n"
"        local val = reflection_get(err, getter)\n"
"        if val ~= nil then\n"
"            return val\n"
"        end\n"
"    end\n"
"    return error_methods[key]\n"
"end\n"
"\n"
"local error_mt = {\n"
"    __index = error_index;\n"
"    __tostring = error_message;\n"
"};\n"
"\n"
"ffi.metatype('struct error', error_mt);\n"
"\n"
"return error\n"
""
;
